/*
  proxy.c : anti firewall proxy

    Copyright (C) 1999 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>

#include <unistd.h>
#include <fcntl.h>

#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

static int debug = 0;

static int get_arg_port(int argc, char *argv[])
{
	int port, n;

	port = -1;
	for (n = 1; n < argc; n++) {
	    if (strcmp(argv[n], "-d") == 0) {
		    debug = 1;
	    } else if (strcmp(argv[n], "-p") == 0 && n < argc) {
		    /* specify port we listen in */
		    n++;
		    port = atoi(argv[n]);
	    } else {
		    printf("Invalid argument %s\n", argv[n]);
		    port = -1;
		    break;
	    }
	}

	if (port == -1) {
		printf("Usage: [-d] -p <port to listen>\n");
		return -1;
	}

	return port;
}

static int net_listen(int port)
{
	struct sockaddr_in sin;
	int handle;

	handle = socket(AF_INET, SOCK_STREAM, 0);
	if (handle == -1) {
		fprintf(stderr, "socket() failed: %s\n", strerror(errno));
		return -1;
	}

	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_port = htons(port);
	if (bind(handle, (struct sockaddr *) &sin, sizeof(sin)) == -1) {
		printf("bind() failed: %s\n", strerror(errno));
		return -1;
	}

	if (listen(handle, 1) == -1) {
		printf("listen() in port %d failed: %s\n", port, strerror(errno));
		return -1;
	}

	fcntl(handle, F_SETFL, O_NONBLOCK);
	return handle;
}

int main(int argc, char *argv[])
{
	struct sockaddr_in saddr;
	fd_set set;

	char buf[1024];
	int listenh, clienth, serverh, handle;
	int max, port, len, ret;

	port = get_arg_port(argc, argv);
	if (port == -1) return 1;

	listenh = net_listen(port);
	if (listenh == -1) return 1;

	if (!debug) {
		switch (fork()) {
		case -1:
			printf("fork() failed\n");
			return 1;
		case 0:
			break;
		default:
			return 0;
		}
	}

        serverh = clienth = -1;
	for (;;) {
		FD_ZERO(&set);

		max = listenh;
		FD_SET(listenh, &set);
		if (serverh != -1) {
			FD_SET(serverh, &set);
			if (max < serverh) max = serverh;
		}
		if (clienth != -1) {
			FD_SET(clienth, &set);
			if (max < clienth) max = clienth;
		}
		if (!select(max+1, &set, NULL, NULL, NULL)) {
			if (clienth == -1 && serverh != -1)
				send(serverh, "P", 1, 0);
			continue;
		}

		if (FD_ISSET(listenh, &set)) {
                        /* server or client connected */
			len = sizeof(saddr);
			handle = accept(listenh, (struct sockaddr *) &saddr, &len);
			if (serverh == -1) {
				serverh = handle;
				fprintf(stderr, "server connected\n");
			} else if (clienth == -1) {
				/* client connected - send a byte to server
				   so that it will know that connection has
				   been opened */
                                if (send(serverh, "N", 1, 0) == 1)
					clienth = handle;
				else {
					fprintf(stderr, "accept send() failed: %s\n", strerror(errno));
					close(handle);
				}
			} else {
				/* third connection? nope, don't allow. */
				close(handle);
			}
		}

		if (serverh != -1 && FD_ISSET(serverh, &set)) {
		        /* data from server - send to client */
			ret = recv(serverh, buf, sizeof(buf), 0);
			if (ret > 0) {
				if (clienth == -1)
					continue;

				if (send(clienth, buf, ret, 0) != ret) {
					fprintf(stderr, "client send() failed: %s\n", strerror(errno));
					close(serverh); serverh = -1;
					close(clienth); clienth = -1;
				}
			} else if (ret == 0 || (errno != EAGAIN && errno != EWOULDBLOCK)) {
				fprintf(stderr, "server recv() failed: %s\n", strerror(errno));
				close(serverh); serverh = -1;
				close(clienth); clienth = -1;
			}
		}

		if (clienth != -1 && FD_ISSET(clienth, &set)) {
			ret = recv(clienth, buf, sizeof(buf), 0);
			if (ret > 0) {
				if (send(serverh, buf, ret, 0) != ret) {
					fprintf(stderr, "server send() failed: %s\n", strerror(errno));
					close(serverh); serverh = -1;
					close(clienth); clienth = -1;
				}
			} else if (ret == 0 || (errno != EAGAIN && errno != EWOULDBLOCK)) {
				fprintf(stderr, "client recv() failed: %s\n", strerror(errno));
				close(serverh); serverh = -1;
				close(clienth); clienth = -1;
			}
		}
	}

	return 0;
}
