/*
  server.c : anti firewall server

    Copyright (C) 1999 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>

#include <unistd.h>
#include <fcntl.h>

#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

static int recon_wait;

static int get_args(int argc, char *argv[], char **outaddr)
{
	int n, port;

	port = -1;
	*outaddr = NULL;

	for (n = 1; n < argc; n++) {
		if (strcmp(argv[n], "-h") == 0 && n < argc) {
			/* specify address where we connect to */
			n++;
			*outaddr = argv[n];
		}
		else if (strcmp(argv[n], "-p") == 0 && n < argc) {
			/* specify port we connect to */
			n++;
			port = atoi(argv[n]);
		} else {
			printf("Invalid argument %s\n", argv[n]);
			*outaddr = NULL;
			break;
		}
	}

	return port;
}

static int net_connect(char *addr, int port)
{
	struct hostent *hp;
	struct sockaddr_in sin;
	int handle;

	hp = gethostbyname(addr);
	if (hp == NULL) return -1;

	handle = socket(AF_INET, SOCK_STREAM, 0);
	if (handle == -1)
		return -1;

	/* connect */
	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	memcpy(&sin.sin_addr, hp->h_addr, 4);
	sin.sin_port = htons(port);

	if (connect(handle, (struct sockaddr *) &sin, sizeof(sin)) == -1) {
		close(handle);
		return -1;
	}

	fcntl(handle, F_SETFL, O_NONBLOCK);
	return handle;
}

static void server_connect(char *addr, int port, char *redir_addr, int redir_port)
{
	struct timeval tv;
	fd_set set;

	char buf[1024];
	int proxyh, clienth, ret;
	time_t last, idle;

	/* connect to proxy */
	proxyh = net_connect(addr, port);
	if (proxyh == -1) {
		if (recon_wait < 15) recon_wait++;
		sleep(60*recon_wait);
		return;
	}

	recon_wait = 0;
	clienth = -1;
	idle = time(NULL);

	last = time(NULL);
	for (;;) {
		if (clienth == -1 && time(NULL)-last >= 60*30) {
			/* reconnect every 30min, just in case if
			   something breaks */
			break;
		}

		tv.tv_sec = 10;
		tv.tv_usec = 0;

		FD_ZERO(&set);
		FD_SET(proxyh, &set);
		if (clienth != -1) FD_SET(clienth, &set);
		if (!select((proxyh > clienth ? proxyh : clienth)+1,
			    &set, NULL, NULL, &tv)) {
			if (clienth == -1) {
				if (send(proxyh, "P", 1, 0) != 1)
					break;
			} else if (time(NULL) > idle+60*30) {
				/* 30 min idle, disconnect */
				break;
			}
			continue;
		}
		idle = time(NULL);

		if (FD_ISSET(proxyh, &set) && clienth == -1) {
			/* client connected to proxy - redirect the
			   connection */
			if (recv(proxyh, buf, 1, 0) == 1 && buf[0] == 'N') {
				clienth = net_connect(redir_addr, redir_port);
				if (clienth == -1) break;
			}
			continue;
		}

		if (clienth != -1 && FD_ISSET(clienth, &set)) {
			ret = recv(clienth, buf, sizeof(buf), 0);
			if (ret == 0 || (ret == -1 && errno != EAGAIN && errno != EWOULDBLOCK))
				break;

			send(proxyh, buf, ret, 0);
		}

		if (FD_ISSET(proxyh, &set)) {
			ret = recv(proxyh, buf, sizeof(buf), 0);
			if (ret == 0 || (ret == -1 && errno != EAGAIN && errno != EWOULDBLOCK))
				break;

 			send(clienth, buf, ret, 0);
		}
	}

	if (clienth != -1) close(clienth);
	close(proxyh);
}

int main(int argc, char *argv[])
{
	char *outaddr;
	int outport;

	outport = get_args(argc, argv, &outaddr);
	if (outaddr == NULL || outport == -1) {
		printf("Usage: -h <address to connect> -p <port to connect>\n");
		return 1;
	}

	switch (fork()) {
	case -1:
		printf("fork() failed\n");
		return 1;
	case 0:
		break;
	default:
		return 0;
	}

	recon_wait = 0;
	for (;;) {
		server_connect(outaddr, outport, "127.0.0.1", 22);
		sleep(1);
	}

	return 0;
}
